program PBA;

{
  Pascal Builder for Android.

  This Free Pascal console application is run as a "builder" from Eclipse
   and does the following for an Eclipse Android project:

   - Creates each activity's .pas file if it doesn't exist yet.

   - Parses R.java resource file and creates Pascal equivalent (Rjava.pas).

   - Runs FPC's JVM compiler to compile project's main .pas file.

   - Deletes .ppu files from classes directory so Android Package Manager
     doesn't include .ppu files in .apk package.

  Author:     Phil Hess.
  Copyright:  Copyright (C) 2012 Phil Hess. All rights reserved.
  License:    Modified LGPL.
}

{$MODE Delphi}

uses
  SysUtils,
  DOM,
  XmlRead,
  Process; 

const
  ProgName     = 'pba';

var
  FpcPath      : string;
  FpcRtlDir1   : string;
  FpcRtlDir2   : string;
  ProjDir      : string;
  ManifestPath : string;
  Namespace    : string;
  ActivityName : string;
  LayoutName   : string;
  PasPath      : string;
  MainPasPath  : string;
  RJavaPath    : string;
  RPasPath     : string;
  ClassesDir   : string;

  ManifestXml  : TXMLDocument;
  AttrNode     : TDOMNode;
  AppNode      : TDOMNode;
  CurNode      : TDOMNode;
  NodeIdx      : Integer;

  InF          : TextFile;
  OutF         : TextFile;
  InStr        : string;
  InClass      : Boolean;
  InConst      : Boolean;
  AProcess     : TProcess;
  ParamIdx     : Integer;
  SrchResult   : Integer;
  SrchRec      : TSearchRec;

const
{$IFDEF MSWINDOWS}
  NewLine = #13#10;
{$ELSE}
  NewLine = #10;
{$ENDIF}

begin

  if ParamCount < 3 then
    begin
    WriteLn(stderr, 'Pascal Builder for Android - this app is run as a "builder" from Eclipse.');
    WriteLn(stderr, 'Arguments: path/to/ppcjvm path/to/Android/project_dir -FDpath/to/jasmin_dir [FPC switches]');
    WriteLn(stderr, 'Example:   ~/Tools/fpcjvm/lib/fpc/2.7.1/ppcjvm ${project_loc} -FD/usr/local/bin');
    Exit;
    end;

   {Check various paths}
  FpcPath := ExpandFileName(ParamStr(1));
  if not FileExists(FpcPath) then
    begin
    WriteLn(stderr, ParamStr(1));
    WriteLn(stderr, 'File does not exist - specify path to FPC JVM compiler as first argument.');
    Exit;
    end;

  FpcRtlDir1 := IncludeTrailingPathDelimiter(ExtractFileDir(FpcPath)) +
                'units' + PathDelim + 'jvm-android' + PathDelim + 'rtl';
  FpcRtlDir2 := FpcRtlDir1 + PathDelim + 'org' + PathDelim + 'freepascal' + 
                PathDelim + 'rtl';

  ProjDir := ExcludeTrailingPathDelimiter(ExpandFileName(ParamStr(2)));
  if not DirectoryExists(ProjDir) then
    begin
    WriteLn(stderr, ParamStr(2));
    WriteLn(stderr, 'Directory does not exist - specify ${project_loc} as second argument.');
    Exit;
    end;

  ManifestPath := ProjDir + PathDelim + 'AndroidManifest.xml';
  if not FileExists(ManifestPath) then
    begin
    WriteLn(stderr, ManifestPath);
    WriteLn(stderr, 'Manifest file is missing from Android project directory.');
    Exit;
    end;


   {Load manifest file and extract package name to use as namespace.
    Check each activity and create its .pas file if it doesn't yet exist.}
  ManifestXml := TXMLDocument.Create;
  try
    ReadXmlFile(ManifestXml, ManifestPath);
    if ManifestXml.DocumentElement.NodeName <> 'manifest' then
      begin
      WriteLn(stderr, ManifestPath);
      WriteLn(stderr, 'Does not appear to be a valid Android manifest file.');
      Exit;
      end;
    AttrNode := ManifestXml.DocumentElement.Attributes.GetNamedItem('package');
    Namespace := AttrNode.NodeValue;

    MainPasPath := '';
    AppNode := ManifestXml.DocumentElement.FindNode('application');
    for NodeIdx := 0 to AppNode.ChildNodes.Count-1 do
      begin
      CurNode := AppNode.ChildNodes.Item[NodeIdx];
      if CurNode.NodeName = 'activity' then
        begin
        AttrNode := CurNode.Attributes.GetNamedItem('android:name');
        ActivityName := AttrNode.NodeValue;
        ActivityName := Copy(ActivityName, 2, MaxInt);  {Strip leading dot}
        AttrNode := CurNode.Attributes.GetNamedItem('android:label');
        LayoutName := AttrNode.NodeValue;
        LayoutName := Copy(LayoutName, 15, MaxInt);  {Strip "@string/title_"}

        PasPath := ProjDir + PathDelim + 'pas' + PathDelim + ActivityName + '.pas';
        if MainPasPath = '' then
          MainPasPath := PasPath;  //assume first one is main activity

         {Create Pascal file from boilerplate code if it doesn't exist.
          Note not creating under 'src' or else .pas files get copied 
           for some reason to classes directory and thus into .apk.}
        if not FileExists(PasPath) then
          begin
          ForceDirectories(ExtractFileDir(PasPath));
          AssignFile(OutF, PasPath);
          Rewrite(OutF);
          WriteLn(OutF,
'unit ' + ActivityName + ';' + NewLine +
NewLine +
'{$mode delphi}' + NewLine +
'{$modeswitch unicodestrings}' + NewLine +
'{$namespace ' + Namespace + '}' + NewLine +
NewLine +
'interface' + NewLine +
NewLine +
'uses androidr14, Rjava;' + NewLine +
NewLine +
'type' + NewLine +
'  ' + ActivityName + ' = class(AAActivity)' + NewLine +
'    procedure onCreate(savedInstanceState: AOBundle); override;' + NewLine +
'    function onCreateOptionsMenu(menu: AVMenu): jboolean; override;' + NewLine +
'  end;' + NewLine +
NewLine +
NewLine +
'implementation' + NewLine +
NewLine +
'procedure ' + ActivityName + '.onCreate(savedInstanceState: AOBundle);' + NewLine +
'begin' + NewLine +
'  inherited onCreate(savedInstanceState);' + NewLine +
'  setContentView(R.layout.' + LayoutName + ');' + NewLine +
'end;' + NewLine +
NewLine +
'function ' + ActivityName + '.onCreateOptionsMenu(menu: AVMenu): jboolean;' + NewLine +
'begin' + NewLine +
'  getMenuInflater().inflate(R.menu.' + LayoutName + ', menu);' + NewLine +
'  Result := True;' + NewLine +
'end;' + NewLine +
NewLine +
'end.' + NewLine + NewLine);

          CloseFile(OutF);
          end;
        end;      
      end;
  finally
    ManifestXml.Free;
  end;


   {Parse R.java file and create Rjava.pas equivalent if R.java is newer
     than Rjava.pas.
    Note not naming file R.pas or unit will have same name as class,
     which would require R.R. instead of R. to reference constants.}
  RJavaPath := ProjDir + PathDelim + 'gen' + PathDelim +
               StringReplace(Namespace, '.', PathDelim, [rfReplaceAll]) + 
               PathDelim + 'R.java';
  if not FileExists(RJavaPath) then
    begin
    WriteLn(stderr, RJavaPath);
    WriteLn(stderr, 'File does not exist.');
    Exit;
    end;

  RPasPath := ProjDir + PathDelim + 'pas' + PathDelim + 'Rjava.pas';
  if (not FileExists(RPasPath)) or
     (FileAge(RPasPath) < FileAge(RJavaPath)) then
    begin
    AssignFile(InF, RJavaPath);
    Reset(InF);
    AssignFile(OutF, RPasPath);
    Rewrite(OutF);
  
    WriteLn(OutF,
'// AUTO-GENERATED FILE. DO NOT MODIFY.' + NewLine +
NewLine +
'// This class was automatically generated by the ' + ProgName + ' tool' + NewLine +
'// from R.java. It should not be modified by hand.' + NewLine +
NewLine +
'unit Rjava;' + NewLine +
NewLine +
'{$mode delphi}' + NewLine +
'{$modeswitch unicodestrings}' + NewLine +
'{$namespace ', Namespace, '}' + NewLine +
NewLine +
'interface' + NewLine +
NewLine +
'type' + NewLine +
'  R = class' + NewLine +
'  public' + NewLine +
'    type');

    repeat
      ReadLn(InF, InStr);
      InStr := Trim(InStr);
    until Copy(InStr, 1, 21) = 'public final class R ';
    InClass := False;
    InConst := False;
    while not Eof(InF) do
      begin
      ReadLn(InF, InStr);
      InStr := Trim(InStr);
      if Copy(InStr, 1, 26) = 'public static final class ' then
        begin
        InStr := Copy(InStr, 27, MaxInt);
        if Pos('{', InStr) > 0 then
          InStr := Copy(InStr, 1, Pos('{', InStr)-1);
        InStr := Trim(InStr);
        if InStr = 'string' then
          InStr := 'string_';
        WriteLn(OutF, '      ', InStr, ' = class');
        InClass := True;
        end
      else if InClass and (Copy(InStr, 1, 24) = 'public static final int ') then
        begin
        if not InConst then
          begin
          WriteLn(OutF, '      public');
          WriteLn(OutF, '        const');
          InConst := True;
          end;
        InStr := Copy(InStr, 25, MaxInt);
        InStr := StringReplace(InStr, '=0x', ' = $', []);
        WriteLn(OutF, '          ', InStr);
        end
      else if InClass and (InStr = '}') then
        begin
        WriteLn(OutF, '      end;');
        WriteLn(OutF);
        InClass := False;
        InConst := False;
        end;
      end;
    WriteLn(OutF, '  end;');
    WriteLn(OutF);
    WriteLn(OutF, 'implementation');
    WriteLn(OutF);
    WriteLn(OutF, 'end.');
    CloseFile(InF);
    CloseFile(OutF);
    end;


   {Launch FPC compiler}
  ClassesDir := ProjDir + PathDelim + 'bin' + PathDelim + 'classes';

  AProcess := TProcess.Create(nil);
  AProcess.Executable := FpcPath;
  AProcess.Parameters.Add('-Tandroid');
  AProcess.Parameters.Add('-O2');
  AProcess.Parameters.Add('-g');
  AProcess.Parameters.Add('-XP');
  AProcess.Parameters.Add('-Fu' + FpcRtlDir1);
  AProcess.Parameters.Add('-Fu' + FpcRtlDir2);
  AProcess.Parameters.Add('-FU' + ClassesDir);
  for ParamIdx := 3 to ParamCount do  {Add any additional params (-FD, etc.)}
    AProcess.Parameters.Add(ParamStr(ParamIdx));
  AProcess.Parameters.Add(MainPasPath);
  AProcess.Options := [poUsePipes];

  try
    AProcess.Execute;
    repeat
      if AProcess.Output.NumBytesAvailable > 0 then
        begin
        SetLength(InStr, AProcess.Output.NumBytesAvailable);
        AProcess.Output.Read(InStr[1], Length(InStr));   
        Write(InStr); 
        end;
    until not AProcess.Running;
  finally
    AProcess.Free;
  end;


   {Delete .ppu files from project's classes directory so not added to .apk}
  SrchResult := FindFirst(ClassesDir + PathDelim + '*.ppu', 0, SrchRec);
  while SrchResult = 0 do
    begin
    DeleteFile(ClassesDir + PathDelim + SrchRec.Name);
    SrchResult := FindNext(SrchRec);
    end;
  FindClose(SrchRec);

end.

